import os
import logging
import pathlib

import falcon


logger = logging.getLogger(__name__)


class StaticsMiddleware:
    """Statics Middleware

    Return static files from disk with a matching route.

    Example:
        import falcon
        from falcon_helpers.middlewares import statics

        api = falcon.api(
            middlewares=[statics.StaticsMiddleware()]
        )

        # HTTP GET /static/js/main.js

    Attributes:
        static_url (string): The url to look for when processing the request.
            This route is hijacked by the middleware, so any request going to it
            will attempt to find the remaining url path on the file system.

        static_folder (string): the folder, relative to the current working
            directory which contains the static files. Any attempts to go up a
            directory is stripped out (../), see ``clean_path()``

        content_types (dict): a dict containing additional file type to content
            type mappings.
    """

    extension_content_types = {
        'css': 'text/css',
        'js': 'application/javascript',
        'html': 'text/html',
        'svg': 'image/svg+xml'
    }

    def __init__(self, static_url='/static', static_folder='static',
                 content_types=None):
        self.static_url = static_url

        if content_types:
            self.extension_content_types.update(content_types)

        # Use the current working directory to search for the static folder
        path = pathlib.Path(os.getcwd())
        self.static_folder = path.joinpath(static_folder).resolve()
        logger.info('Using {self.static_folder} for the static.')

    def process_request(self, req, resp):
        prefixed = req.path.startswith(self.static_url)

        if not prefixed:
            return

        stripped = req.path.replace(self.static_url, '')
        cleaned = self.clean_path(stripped)
        path = self.static_folder.joinpath(cleaned).resolve()

        try:
            resp.stream = open(path, 'rb')
        except FileNotFoundError:
            raise falcon.HTTPNotFound()

        resp.content_type = self.extension_content_types.get(
            path.suffix.lstrip('.'), 'application/octet-stream')

        raise falcon.HTTPStatus('200 OK')

    def clean_path(self, path):
        """Cleanup a path to prevent traversing any directory above the
        static folder
        """
        return path.replace('../', '').lstrip('/')



import falcon
import jwt

def _default_failed(exception):
    raise falcon.HTTPUnauthorized()


class ConfigurationError(Exception):
    pass


class ParseJWTMiddleware:
    """Strip a JWT from a cookie or a header

    Example:
        import falcon
        from falcon_helpers.middlewares.parsejwt import ParseJWTMiddleware

        api = falcon.API(
            middleware=[
                ParseJWTMiddleware(
                    audience='your-audience',
                    secret='a-really-great-random-string',
                    cookie_name='MyCookieName')
            ]
        )

    Attributes:
        audience: (string) A string audience which is passed to the JWT decoder

        secret: (string) A secret key to verify the token

        pubkey: (string) When using RS256, pass a public key not a token

        when_fails: (function) A function to execute when the authentication
            fails

        cookie_name: (string) the name of the cookie to look for

        header_name: (string) the name of the cookie to look for

        context_key: (string) the key to put the JWT

    """

    def __init__(self, audience, secret=None, pubkey=None,
                 cookie_name=None, header_name=None,
                 context_key='auth_token_contents',
                 when_fails=_default_failed):

        if cookie_name and header_name:
            raise ConfigurationError('Can\'t pull the token from both a header'
                                     ' and a cookie')

        if not cookie_name and not header_name:
            cookie_name = 'X-AuthToken'

        self.audience = audience
        self.secret = secret
        self.pubkey = pubkey
        self.cookie_name = cookie_name
        self.header_name = header_name
        self.failed_action = when_fails

        if (self.pubkey is not None
            and not self.pubkey.startswith('ssh-rsa')):
            raise ConfigurationError(
                'A public key for RS256 encoding must be in PEM Format')

    def verify_request(self, token):
        if not token:
            raise ValueError('No token found')

        header = jwt.get_unverified_header(token)

        (type_, verify_with) = (('public key', self.pubkey)
                                if header['alg'] == 'RS256'
                                else ('secret key', self.secret))

        if verify_with is None:
            raise ConfigurationError('You must pass the correct verification'
                                     f' type for this algorithm.'
                                     f' {header["alg"]} requires a {type_}.')

        return jwt.decode(token, verify_with, audience=self.audience)


    def process_request(self, req, resp):
        token = req.cookies.get(self.cookie_name, False)

        try:
            req.context['auth_token_contents'] = self.verify_request(token)
        except Exception as e:
            return self.failed_action(e, req, resp)



import falcon
import requests

from ..middlewares.jinja2 import Jinja2Response


class LoginFormResource:
    auth_required = False

    def __init__(self, client_url, client_id, callback_uri):
        self.client_url = client_url
        self.client_id = client_id
        self.callback_uri = callback_uri

    # Load form
    def on_get(self, req, resp):
        resp.context['template'] = Jinja2Response(
            'auth/login.html',
            auth0_config={
                'client_url': self.client_url,
                'client_id': self.client_id,
                'callback_uri': self.callback_uri,
            })


class LoginCallbackResource:
    auth_required = False

    def __init__(self, client_url, client_id, client_secret, callback_uri,
                 cookie_domain, secure_cookie, cookie_name='X-AuthToken',
                 cookie_max_age=(24 * 60 * 60), cookie_path='/',
                 redirect_uri='/dashboard'
                 ):
        self.client_url = client_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.callback_uri = callback_uri
        self.redirect_uri = redirect_uri

        self.cookie_domain = cookie_domain
        self.secure_cookie = secure_cookie
        self.cookie_name = cookie_name
        self.cookie_max_age = cookie_max_age
        self.cookie_path = cookie_path

    def on_get(self, req, resp):
        try:
            auth_code = req.params['code']
        except KeyError:
            raise falcon.HTTPForbidden('Forbidden', 'Missing access code')

        data = self.get_user_data(auth_code)

        self.set_cookie_contents(resp, data)
        resp.set_header('Location', self.redirect_uri)
        resp.status = falcon.HTTP_302

    def get_user_data(self, auth_code):
        token = requests.post(
            f'https://{self.client_url}/oauth/token',
            data={
                'grant_type': 'authorization_code',
                'client_id': self.client_id,
                'client_secret': self.client_secret,
                'code': auth_code,
                'redirect_uri': self.callback_uri,
            })

        if not token.ok:
            raise falcon.HTTPTemporaryRedirect('/auth/login')

        return token.json()

    def set_cookie_contents(self, resp, data):
        resp.set_cookie(self.cookie_name,
                        data['id_token'],
                        max_age=self.cookie_max_age,
                        secure=self.secure_cookie,
                        path=self.cookie_path,
                        domain=self.cookie_domain)


class LogoutResource:
    auth_required = False

    def __init__(self, cookie_domain, secure_cookie, cookie_name='X-AuthToken',
                 cookie_path='/'):
        self.cookie_domain = cookie_domain
        self.secure_cookie = secure_cookie

        self.cookie_name = cookie_name
        self.cookie_path = cookie_path

    def on_get(self, req, resp):
        resp.set_cookie(self.cookie_name, '', secure=self.secure_cookie,
                        path=self.cookie_path, domain=self.cookie_domain)

        resp.status = falcon.HTTP_302
        resp.set_header('Location', '/auth/login')



import ast
import pathlib
import sqlalchemy as sa
import sqlalchemy.types as types


class CartesianPoint:
    def __init__(self, point):
        self._raw = point
        self.x = point[0]
        self.y = point[1]

    def __getitem__(self, key):
        return self._raw[key]

    def __str__(self):
        return self._raw


class Path(types.TypeDecorator):
    """Automatically coerces a pathlib.Path to Unicode and back again"""

    impl = types.Unicode

    def process_bind_param(self, value, dialect):

        if isinstance(value, (pathlib.PurePath, str)):
            return str(value)
        elif value is None:
            return None
        else:
            raise TypeError(f'Unable to store {type(value)}')

    def process_result_value(self, value, dialect):
        if value:
            return pathlib.Path(value)

    def copy(self, **kw):
        return Path(self.impl.length)


class Point(types.UserDefinedType):

    def get_col_spec(self):
        return 'point'

    def bind_processor(self, dialect):
        def process(value):
            # Just in case this is null, we should allow it, the DB will cath
            # nullability
            if value is None:
                return None

            return str(value)
        return process

    def result_processor(self, dialect, coltype):
        def process(value):
            if not value:
                return None

            v = ast.literal_eval(value)
            return CartesianPoint(v)
        return process



import falcon


class RedirectResource:
    """A simple redirect resource which can be used to force a route to
    redirect.

    :params path: the string version of the path to redirect to
    :params status: a falcon status object, default to falcon.HTTP_301
    """

    def __init__(self, path, status=None):
        self.path = path
        self.status = status or falcon.HTTP_301

    def on_get(self, req, resp):
        resp.status = self.status
        resp.set_header('Location', self.path)



import io
import os.path
import pathlib
import mimetypes

try:
    import botocore
    import boto3
except ImportError:
    pass

import uuid

import falcon


class Document:
    def __init__(self, name, uid, path, storage_type=None, **kwargs):
        self.name = name
        self.uid = uid
        self.path = path
        self.storage_type = storage_type
        self.details = kwargs


class StorageException(Exception):
    pass


class S3FileStore:
    storage_type = 'OBJECT::S3'

    def __init__(self, bucket, prefix='/', uidgen=uuid.uuid4,
                 default_content_type='application/octet-stream',
                 endpoint=None
                 ):
        if not prefix.startswith('/'):
            raise AssertionError('S3ImageStore requires an absolute path.')

        # Setup the mimetypes database
        mimetypes.init()

        # Remove any trailing slash
        self.prefix = prefix.rstrip('/') if prefix != '/' else prefix
        self.bucket = bucket
        self.uidgen = uidgen
        self.default_content_type = default_content_type
        self.endpoint = endpoint

    @property
    def connection(self):
        return boto3.resource('s3')

    @property
    def client(self):
        return boto3.client('s3', endpoint_url=self.endpoint)

    def make_download_url(self, doc, expires_in=60):
        return self.client.generate_presigned_url(
            ClientMethod='get_object',
            ExpiresIn=expires_in,
            Params={
                'Bucket': self.bucket,
                'Key': doc.path.lstrip('/'),
                'ResponseContentDisposition': 'attachment; filename="{}"'.format(doc.name)
            })

    def make_response(self, doc, resp):
        resp.status = falcon.HTTP_303
        resp.location = self.make_download_url(doc)

    def save(self, filename, fp, path=None):
        unique_name = self.uidgen()
        (content_type, content_encoding) = mimetypes.guess_type(filename)

        if not content_type:
            content_type = self.default_content_type

        name = unique_name

        if path:
            final_path = self.prefix + '/' + path.strip('/') + '/' + str(name)
        else:
            final_path = self.prefix + '/' + str(name)

        final_path = os.path.normpath(final_path)

        if not final_path.startswith(self.prefix):
            raise StorageException('Invalid path')

        # This could raise an exception, catch it at the app level so you have the reporting
        # capabilities... if only python had a Result type...
        obj = self.connection.Object(self.bucket, final_path.lstrip('/'))
        result = obj.put(Body=fp, ContentType=content_type, Metadata={ 'original-filename': filename })

        if ('ResponseMetadata' in result and
            result['ResponseMetadata'].get('HTTPStatusCode', None) in (200, 201)):

            return Document(uid=unique_name, name=filename, path=final_path,
                            storage_type=self.storage_type, response=result)
        else:
            raise StorageException('Unknown State. Response Type Unknown.')


class LocalFileStore:
    storage_type = 'FILE::LOCAL'

    def __init__(self, path, uidgen=uuid.uuid4, default_content_type='application/octet-stream',
                 _fopen=io.open):
        self.path = pathlib.Path(path)
        self.uidgen = uidgen
        self._fopen = _fopen

        # Setup the mimetypes database
        mimetypes.init()
        self.default_content_type = default_content_type

    def save(self, filename, fp, path=None):
        unique_name = self.uidgen()

        parent = self.path.joinpath(path)

        parent.mkdir(parents=True, exist_ok=True)
        final_path = parent.joinpath(str(unique_name))

        with open(final_path, 'wb') as f:
            result = f.write(fp.read())

        return Document(uid=unique_name,
                        name=filename,
                        path=str(final_path),
                        storage_type=self.storage_type,
                        response=result)

    def make_response(self, doc, resp):
        (content_type, content_encoding) = mimetypes.guess_type(doc.name)

        if not content_type:
            content_type = self.default_content_type

        resp.stream = self._fopen(doc.path, 'rb')
        stream_len = os.path.getsize(doc.path)
        resp.content_length = stream_len

        resp.downloadable_as = doc.name

        resp.content_type = content_type

        resp.status = falcon.HTTP_200



import random
import sqlalchemy as sa
from sqlalchemy.schema import MetaData
from sqlalchemy.ext.declarative import as_declarative

from falcon_helpers.sqla.core import utcnow
from falcon_helpers.sqla.utils import random_data_for_type


convention = {
  "ix": 'ix_%(column_0_label)s',
  "uq": "uq_%(table_name)s_%(column_0_name)s",
  "ck": "ck_%(table_name)s_%(constraint_name)s",
  "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
  "pk": "pk_%(table_name)s"
}

metadata = MetaData(naming_convention=convention)


class BaseColumns:
    id = sa.Column(sa.Integer, primary_key=True, nullable=False)

    created_ts = sa.Column(sa.DateTime, server_default=utcnow())
    updated_ts = sa.Column(sa.DateTime, server_default=utcnow(),
                           server_onupdate=utcnow())


class Testable:
    testing_random_nulls = True

    @classmethod
    def testing_create(cls, **kwargs):
        """Create an object for testing with default data appropriate for the
        field type


        :param _numeric_defaults_range: a tuple of (HIGH, LOW) which controls
            the acceptable defaults of the number types. Both integer and
            numeric (float) fields are controlled by this setting. This is
            helpful when some fields have a constrained value.

        ATTRIBUTION: This was largely copied from the wonderful folks at Level
        12, they write high-class software, check them out: https://level12.io
        """

        NUMERIC_LOW, NUMERIC_HIGH = kwargs.pop('_numeric_defaults_range', (-100, 100))

        insp = sa.inspection.inspect(cls)

        def skippable(column):
            return (
                column.key in kwargs     # skip fields already in kwargs
                or column.foreign_keys   # skip foreign keys
                or column.server_default # skip fields with server defaults
                or column.default        # skip fields with defaults
                or column.primary_key)   # skip any primary key

        for column in (col for col in insp.columns if not skippable(col)):
            try:
                if column.nullable and cls.testing_random_nulls and random.choice([True, False]):
                    kwargs[column.key] = None
                else:
                    kwargs[column.key] = random_data_for_type(
                        column, NUMERIC_HIGH, NUMERIC_LOW)
            except ValueError:
                pass

        return cls(**kwargs)


@as_declarative(metadata=metadata)
class ModelBase(Testable):
    pass



from sqlalchemy.orm import (
    scoped_session,
    sessionmaker,
)

from falcon_helpers.sqla.db import session


class SQLAlchemySessionMiddleware:
    def __init__(self, session):
        self.session = session

    def process_resource(self, req, resp, resource, params):
        resource.session = self.session()

    def process_response(self, req, resp, resource, req_succeeded):
        if not hasattr(resource, 'session'):
            return

        try:
            if not req_succeeded:
                resource.session.rollback()
            else:
                resource.session.commit()
        except:
            resource.session.close()



import os
import logging

import jinja2
from jinja2 import FileSystemLoader
from jinja2.environment import Environment


logger = logging.getLogger(__name__)

class Jinja2ConfigurationError(Exception):
    pass


class Jinja2Response:
    """A Response Encapsulating a Jinja2 Response

    Utilize this with the Jinja2Middleware to automatically return a rendered
    template as a response.

    Example:
        from falcon_helpers.middlewares import Jinja2Response

        class SomeResource:

            def on_get(self, req, resp):
                data = {
                    "tile": "Some Title"
                }
                resp.context['template'] = Jinja2Response(
                    'path/tp/template',
                    date
                )

    The template will get a few special variables:
        - falcon_request: the current request
        - falcon_response: the current response
        - falcon_resource: the current resource processing the request


    TODO:
        - Add support for globally registered function and variables
    """

    def __init__(self, template, **ctx):
        self.template_fpath = template
        self.ctx = ctx

    def apply(self, resp, env):
        try:
            template = env.get_template(self.template_fpath)
        except jinja2.exceptions.TemplateNotFound as e:
            logger.error(f'Unable to find {self.template_fpath} looking in '
                         f'{env.loader.searchpath}')
            raise


        resp.content_type = 'text/html'
        resp.body = template.render(**self.ctx)


class Jinja2Middleware:

    # TODO: This should probably accept an environment instead of passing around
    # all these one-off environment options
    def __init__(self, template_dpath=None, globals=None):
        """Create a Jinja2Middleware

        :param template_dpath: the path to the Jinja2
        :param globals: a dictionary of globals to attach to the environment

        """
        if not os.path.isabs(template_dpath):
            raise Jinja2ConfigurationError(
                'Template directory path must be absolute')

        self.jinja2_env = Environment()

        # Set up the loader
        self.template_dpath = template_dpath
        self.jinja2_env.loader = FileSystemLoader(self.template_dpath)

        # Add the globals to the env
        self.jinja2_env.globals.update((globals or {}))

    def process_response(self, req, resp, resource, req_succeeded):
        if 'template' not in resp.context:
            return

        resp.context['template'].ctx['falcon_request'] = req
        resp.context['template'].ctx['falcon_response'] = resp
        resp.context['template'].ctx['falcon_resource'] = resource

        resp.context['template'].apply(resp, self.jinja2_env)



import sqlalchemy as sa
import falcon

from ..utils import flatten


class ListBase:
    """A base class for returning list of objects.

    This base class assumes you are using the marshmallow middleware for object
    serialization and sqlalchemy for database access.

    Attributes:
        db_cls: Set this to a SQLAlchemy entity
        schema: Set this a Marshmallow schema
    """
    db_cls = None
    schema = None

    default_order = None

    def on_get(self, req, resp, **kwargs):
        result = self.get_objects(req, **kwargs)
        schema = self.schema(many=True)

        resp.status = falcon.HTTP_200
        resp.body = schema.dump(result)

    def base_query(self, req, **kwargs):
        return self.session.query(self.db_cls)

    def pagination_hook(self, query, req, **kwargs):
        """Create a hook for pagination"""
        size = int(req.params.get('pageSize', 50))

        # -1 here is so that the page numbers start at 1
        page = int(req.params.get('page', 1)) - 1

        if page < 0:
            page = 0

        return query.limit(size).offset((page * size))

    def filter_hook(self, query, req, **kwargs):
        return query

    def order_hook(self, query, req, **kwargs):
        request_order = req.params.get('sort_by', [])

        # This takes advantage of falcons duplicate-param parsing which returns
        # a list when it finds a param multiple times.
        if type(request_order) != list:
            request_order = [request_order]

        request_order = list(flatten([x.split(';') for x in request_order]))
        request_order = [sa.desc(x[1:]) if x.startswith('-') else sa.asc(x) for x in request_order ]

        default_order = request_order or self.default_order or self.db_cls.__mapper__.primary_key
        return query.order_by(*default_order)

    def custom_hook(self, query, req, **kwargs):
        return query

    def get_objects(self, req, *args, **kwargs):
        base = self.base_query(req, **kwargs)
        order = self.order_hook(base, req, **kwargs)
        paged = self.pagination_hook(order, req, **kwargs)
        final = self.custom_hook(paged, req, **kwargs)

        return final.all()


class CrudBase:
    """A very simple CRUD resource.

    This base class assumes you are using the marshmallow middleware for object
    serialization and sqlalchemy for database access.

    Attributes:
        db_cls: Set this to a SQLAlchemy entity
        schema: Set this a Marshmallow schema
    """
    db_cls = None
    schema = None
    default_param_name = 'obj_id'

    def get_object(self, req, **kwargs):
        try:
            obj_id = kwargs[self.default_param_name]
            return self.session.query(self.db_cls).get(obj_id)
        except KeyError:
            raise falcon.HTTPInternalServerError("Misconfigured route")

    def on_get(self, req, resp, **kwargs):
        result = self.get_object(req, **kwargs)

        if not result:
            raise falcon.HTTPNotFound

        schema = self.schema()
        resp.body = schema.dump(result)
        resp.status = falcon.HTTP_200


    def on_put(self, req, resp, **kwargs):
        self.session.add(req.context['dto'].data)
        self.session.flush()

        resp.status = falcon.HTTP_200
        resp.body = self.schema().dump(req.context['dto'].data)


    def on_post(self, req, resp, **kwargs):
        self.session.add(req.context['dto'].data)
        self.session.flush()

        resp.status = falcon.HTTP_201
        resp.body = self.schema().dump(req.context['dto'].data)


    def on_delete(self, req, resp, **kwargs):
        try:
            obj = self.get_object(req, **kwargs)

            if obj:
                obj.delete(synchronize_session=False)

        except sa.exc.IntegrityError:
            self.session.rollback()
            resp.status = falcon.HTTP_400
            resp.json = {'errors': [('Unable to delete because the object is '
                                    'connected to other objects')]}
        else:
            resp.status = falcon.HTTP_204



from .middlewares.statics import StaticsMiddleware
from .middlewares.jinja2 import Jinja2Middleware
from .middlewares.auth_required import AuthRequiredMiddleware

from .resources import auth0

from . import contrib



import falcon


def _default_failed(req, resp, **kwargs):
    raise falcon.HTTPFound('/auth/login')


class AuthRequiredMiddleware:
    """Requires a cookie be set with a valid JWT or fails

    Example:
        import falcon
        from falcon_helpers.middlewares.auth_required import AuthRequiredMiddleware

        class Resource:
            auth_required = True

            def on_get(self, req, resp):
                # ...

        def when_fails_auth(req, resp, token_value):
            raise TerribleException(token_value)

        api = falcon.API(
            middleware=[
                AuthRequiredMiddleware(when_fails=when_fails_auth)
            ]
        )

        api.add_route('/', Resource())

    Attributes:
        resource_param: The paramater to pull the boolean from

        context_key: the key the token will be found on the request

        when_fails: (function) A function to execute when the authentication
            fails
    """

    def __init__(self, resource_param='auth_required',
                 context_key='auth_token_contents',
                 when_fails=_default_failed):

        self.resource_param = resource_param
        self.context_key = context_key
        self.failed_action = when_fails

    def process_resource(self, req, resp, resource, params):
        required = getattr(resource, self.resource_param, True)
        token_value = req.context.get(self.context_key, None)

        token_value = None if isinstance(token_value, Exception) else token_value

        if required and not token_value:
            return self.failed_action(req=req, resp=resp, token_value=token_value)



import falcon
import ujson
from marshmallow.schema import MarshalResult


class MarshmallowMiddleware:

    def process_resource(self, req, resp, resource, params):

        should_parse =  (
            # Veriy that it is safe to parse this resource
            req.method in ('POST', 'PUT'),
            # If there is no data in the body, there is nothing to look at
            bool(req.content_length),
            # If the resource doesn't have a schema the loading would be impossible
            hasattr(resource, 'schema'),
            # Only consider JSON requests for auto-parsing
            req.content_type == 'application/json',
        )

        # Check that all the conditions for parsing are met
        if not all(should_parse):
            req.context['_marshalled'] = False
            return

        stream = req.context['marshalled_stream'] = req.bounded_stream.read()
        data = ujson.loads(stream)

        if req.method == 'PUT':
            data['id'] = params['obj_id']

        schema = resource.schema()
        data = schema.load(data, session=resource.session)

        if data.errors:
            #  This should probably return whatever the accept header indicates
            raise falcon.HTTPStatus(
                falcon.HTTP_400,
                headers={'Content-Type': 'application/json'},
                body=ujson.dumps({'errors': data.errors}),
            )

        req.context['dto'] = data
        req.context['_marshalled'] = True

    def process_response(self, req, resp, resource, req_succeeded):
        if isinstance(resp.body, MarshalResult):
            resp.content_type = 'application/json'
            resp.body = ujson.dumps(resp.body.data)



import enum
import itertools

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr

import falcon
from falcon_helpers.sqla.orm import BaseColumns, metadata, ModelBase
from falcon_helpers.sqla.db import session

from wrapt import decorator

import jwt


user_groups = sa.Table('auth_user_groups', metadata,
    sa.Column('user_id', sa.Integer, sa.ForeignKey('auth_users.id')),
    sa.Column('group_id', sa.Integer, sa.ForeignKey('auth_groups.id'))
)

user_permissions = sa.Table('auth_user_permissions', metadata,
    sa.Column('user_id', sa.Integer, sa.ForeignKey('auth_users.id')),
    sa.Column('permissions_id', sa.Integer, sa.ForeignKey('auth_permissions.id'))
)

group_permissions = sa.Table('auth_group_permissions', metadata,
    sa.Column('group_id', sa.Integer, sa.ForeignKey('auth_groups.id')),
    sa.Column('permissions_id', sa.Integer, sa.ForeignKey('auth_permissions.id'))
)


class User(ModelBase, BaseColumns):
    __tablename__ = 'auth_users'

    ident = sa.Column(sa.Unicode, nullable=False, unique=True)
    is_superuser = sa.Column(sa.types.Boolean, nullable=False, server_default='f')

    groups = sa.orm.relationship('Group', secondary='auth_user_groups')
    assigned_permissions = sa.orm.relationship('Permission', secondary='auth_user_permissions')

    def __repr__(self):
        return self.ident

    def __str__(self):
        return self.ident

    @classmethod
    def get_by_id(cls, ident):
        return sa.orm.Query(cls).filter(cls.ident == ident)

    @property
    def permissions(self):
        if self.is_superuser:
            return session.query(Permission).all()

        group_perms = list(itertools.chain.from_iterable(
            [x.permissions for x in self.groups]))
        user_perms = self.assigned_permissions
        return list(itertools.chain(user_perms, group_perms))

    def has_permission(self, token):
        if self.is_superuser:
            return True

        if not token:
            return False

        # support enum values
        final = token.value if isinstance(token, enum.Enum) else token

        if not isinstance(final, str):
            raise ValueError('Token must be a string when using has_permission')

        return final in self.permissions

    def generate_auth_token(self, audience, secret, algo='HS512'):
        return jwt.encode({
            'sub': self.ident,
            'aud': audience,
        }, secret, algorithm=algo)


class Group(ModelBase, BaseColumns):
    __tablename__ = 'auth_groups'

    ident = sa.Column(sa.Unicode, nullable=False, unique=True)

    def __repr__(self):
        return self.ident

    def __str__(self):
        return self.ident

    @declared_attr
    def permissions(cls):
        return sa.orm.relationship('Permission', secondary='auth_group_permissions')


class Permission(ModelBase, BaseColumns):
    __tablename__ = 'auth_permissions'

    ident = sa.Column(sa.Unicode, nullable=False, unique=True)

    def __repr__(self):
        return '<Permission ident=' + self.ident + '>'

    def __str__(self):
        return self.ident

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.ident

        else:
            super().__eq__(other)


def raise_unauthenticated(*args, **kwargs):
    raise falcon.HTTPUnauthorized()


def route_requires_permission(token=None, on_fail=raise_unauthenticated):
    """Decorate a route to require a certain permission

    This should be used on a falcon resource method such as `on_get`, `on_post`
    to require a base permission for the route. Omitting the token permission
    token requires that the user exists to access that route.

    :param token: the permision token to require
    :param on_fail: a function to execute when the permission check fails.

    NOTE: By default this function will raise an HTTPUnauthorized exception
    """

    @decorator
    def wrapper(wrapped, instance, args, kwargs):
        if (args[0] and
            args[0].context and
            args[0].context.get('user') and
            args[0].context.get('user').has_permission(token)):

            return wrapped(*args, **kwargs)
        else:
            return on_fail(*args, **kwargs)

    return wrapper


def has_permission(user, token):
    return (user and user.has_permission(token))


def has_any_permission(user, tokens):
    return (user and any([user.has_permission(x) for x in tokens]))


def has_all_permissions(user, tokens):
    return (user and all([user.has_permission(x) for x in tokens]))
